package main

import (
	"context"
	"flag"
	"fmt"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	_ "net/http/pprof"
	"os"
	"os/signal"
	"path/filepath"
	"syscall"
	"time"

	"github.com/coreos/go-systemd/v22/daemon"
	"github.com/google/subcommands"
	_ "golang.org/x/net/trace"
	"golang.org/x/sync/errgroup"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials"
	"gopkg.in/yaml.v2"

	"git.autistici.org/ai3/tools/iprep/ext"
	ippb "git.autistici.org/ai3/tools/iprep/proto"
	"git.autistici.org/ai3/tools/iprep/server"
)

type serverCommand struct {
	rpcAddr        string
	httpAddr       string
	dbURI          string
	scriptPath     string
	externalSrcDir string
	tlsCert        string
	tlsKey         string
}

func (c *serverCommand) Name() string     { return "server" }
func (c *serverCommand) Synopsis() string { return "run the rpc server" }
func (c *serverCommand) Usage() string {
	return `server [<flags>]:

        Run the main database and analysis service, with a GRPC API.

`
}

func (c *serverCommand) SetFlags(f *flag.FlagSet) {
	f.StringVar(&c.rpcAddr, "rpc-addr", ":7170", "`address` of GRPC listener")
	f.StringVar(&c.httpAddr, "http-addr", ":7180", "`address` of HTTP debug listener")
	f.StringVar(&c.dbURI, "db", "leveldb:///var/lib/iprep/data", "database `uri` (sqlite:// or leveldb://)")
	f.StringVar(&c.scriptPath, "scoring-script", "/etc/iprep/score.td", "`path` to a custom scoring script")
	f.StringVar(&c.externalSrcDir, "ext-sources-dir", "/etc/iprep/external", "`path` to a directory containing external source definitions")

	f.StringVar(&c.tlsCert, "tls-cert", "", "TLS certificate `path` (grpc only)")
	f.StringVar(&c.tlsKey, "tls-key", "", "TLS private key `path` (grpc only)")
}

func (c *serverCommand) Execute(ctx context.Context, f *flag.FlagSet, args ...interface{}) subcommands.ExitStatus {
	if f.NArg() > 0 {
		log.Printf("error: too many arguments")
		return subcommands.ExitUsageError
	}

	if err := c.run(ctx); err != nil {
		log.Printf("error: %v", err)
		return subcommands.ExitFailure
	}

	return subcommands.ExitSuccess
}

// Definition of an ExternalSource, represented as YAML.
type externalSourceSpec struct {
	Name   string                 `yaml:"name"`
	Type   string                 `yaml:"type"`
	Params map[string]interface{} `yaml:"params"`
}

// Read all files in externalSrcDir and build a map of ExternalSources.
func (c *serverCommand) loadExternalSources() (map[string]ext.ExternalSource, error) {
	files, err := filepath.Glob(filepath.Join(c.externalSrcDir, "*.yml"))
	if err != nil {
		return nil, err
	}

	m := make(map[string]ext.ExternalSource)
	for _, f := range files {
		data, err := ioutil.ReadFile(f)
		if err != nil {
			return nil, err
		}
		var spec externalSourceSpec
		if err := yaml.Unmarshal(data, &spec); err != nil {
			return nil, err
		}
		if spec.Name == "" {
			return nil, fmt.Errorf("error in external source spec %s: missing name", f)
		}
		if _, ok := m[spec.Name]; ok {
			return nil, fmt.Errorf("duplicate external source '%s'", spec.Name)
		}
		src, err := ext.New(spec.Type, spec.Params)
		if err != nil {
			return nil, fmt.Errorf("error creating external source '%s': %v", spec.Name, err)
		}
		m[spec.Name] = src
	}

	return m, nil
}

func (c *serverCommand) run(ctx context.Context) error {
	srcs, err := c.loadExternalSources()
	if err != nil {
		return err
	}
	srv, err := server.New(c.dbURI, c.scriptPath, srcs)
	if err != nil {
		return err
	}
	defer srv.Close()

	var httpServer *http.Server
	var grpcServer *grpc.Server

	var cancel context.CancelFunc
	ctx, cancel = context.WithCancel(ctx)
	g, ctx := errgroup.WithContext(ctx)

	g.Go(func() error {
		httpServer = &http.Server{
			Addr:         c.httpAddr,
			ReadTimeout:  10 * time.Second,
			WriteTimeout: 30 * time.Second,
		}
		log.Printf("starting HTTP server on %s", c.httpAddr)
		if err := httpServer.ListenAndServe(); err != http.ErrServerClosed {
			return err
		}
		return nil
	})

	g.Go(func() error {
		l, err := net.Listen("tcp", c.rpcAddr)
		if err != nil {
			return err
		}
		log.Printf("starting GRPC server on %s", c.rpcAddr)

		var opts []grpc.ServerOption
		if c.tlsCert != "" && c.tlsKey != "" {
			creds, err := credentials.NewServerTLSFromFile(c.tlsCert, c.tlsKey)
			if err != nil {
				return err
			}
			opts = append(opts, grpc.Creds(creds))
		}
		grpcServer = grpc.NewServer(opts...)
		ippb.RegisterIpRepServer(grpcServer, srv)
		return grpcServer.Serve(l)
	})

	// Reload the scoring script on SIGHUP.
	reloadSigCh := make(chan os.Signal, 1)
	go func() {
		for range reloadSigCh {
			srv.Reload()
		}
	}()
	signal.Notify(reloadSigCh, syscall.SIGHUP)

	// Terminate the program on SIGINT/SIGTERM.
	stopSigCh := make(chan os.Signal, 1)
	signal.Notify(stopSigCh, syscall.SIGINT, syscall.SIGTERM)

	// At this point we got nothing else to do but wait for the
	// serving goroutines to return. Notify systemd that we're
	// ready.
	daemon.SdNotify(false, daemon.SdNotifyReady) // nolint

	select {
	case signo := <-stopSigCh:
		log.Printf("terminating due to signal %d", signo)
	case <-ctx.Done():
	}

	// We got here because of an error (or a termination request),
	// so orderly shut down all the serving components.
	cancel()

	if httpServer != nil {
		shutdownCtx, shutdownCancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer shutdownCancel()
		httpServer.Shutdown(shutdownCtx) // nolint
	}
	if grpcServer != nil {
		grpcServer.GracefulStop()
	}

	return g.Wait()
}

func init() {
	subcommands.Register(&serverCommand{}, "")
}
